-- Many of the IDs were verified on live & beta, but some are interpolated.

local MAJOR, MINOR = "LibReforgingInfo-1.0", 3

local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end

--[[
local ReforgingInfo = LibStub("LibReforgingInfo-1.0")
local item = GetInventoryItemLink("player", 1)
if ReforgingInfo:IsItemReforged(item) then
    DEFAULT_CHAT_FRAME:AddMessage(item .. " is reforged:")
    local minus, plus = ReforgingInfo:GetReforgedStatNames(ReforgingInfo:GetReforgeID(item))
    DEFAULT_CHAT_FRAME:AddMessage(minus .. " -> " .. plus)
end
--]]


-- offset is the id of the spirit -> dodge reforge
-- this is the value used in 4.0.3a.
local offset = 113

local SPI = 1
local DODGE = 2
local PARRY = 3
local HIT = 4
local CRIT = 5
local HASTE = 6
local EXP = 7
local MASTERY = 8

-- Array of long stat names. Index = number returned by GetReforgedStatIDs.
lib.StatNames = {
	ITEM_MOD_SPIRIT_SHORT,
	ITEM_MOD_DODGE_RATING_SHORT,
	ITEM_MOD_PARRY_RATING_SHORT,
	ITEM_MOD_HIT_RATING_SHORT,
	ITEM_MOD_CRIT_RATING_SHORT,
	ITEM_MOD_HASTE_RATING_SHORT,
	ITEM_MOD_EXPERTISE_RATING_SHORT,
	ITEM_MOD_MASTERY_RATING_SHORT
}

-- Array of short stat names. Index = number returned by GetReforgedStatIDs.
lib.ShortStatNames = {
	ITEM_MOD_SPIRIT_SHORT,
	STAT_DODGE,
	STAT_PARRY,
	ITEM_MOD_HIT_RATING_SHORT,
	ITEM_MOD_CRIT_RATING_SHORT,
	STAT_HASTE,
	STAT_EXPERTISE,
	STAT_MASTERY
}

-- List of all possible reforgings.
-- First entry has id <offset>, so the order is important!
-- Each entry has the form {decreased stat, increased stat}
local reforgeIDs = {
	{SPI, DODGE},		--113
	{SPI, PARRY},		--114
	{SPI, HIT},			--115
	{SPI, CRIT},		--116
	{SPI, HASTE},		--117
	{SPI, EXP},			--118
	{SPI, MASTERY},		--119
	
	{DODGE, SPI},		--120
	{DODGE, PARRY},		--121
	{DODGE, HIT},		--122
	{DODGE, CRIT},		--123
	{DODGE, HASTE},		--124
	{DODGE, EXP},		--125
	{DODGE, MASTERY},	--126
	
	{PARRY, SPI},		--127
	{PARRY, DODGE},		--128
	{PARRY, HIT},		--129
	{PARRY, CRIT},		--130
	{PARRY, HASTE},		--131
	{PARRY, EXP},		--132
	{PARRY, MASTERY},	--133
	
	{HIT, SPI},			--134
	{HIT, DODGE},		--135
	{HIT, PARRY},		--136
	{HIT, CRIT},		--137
	{HIT, HASTE},		--138
	{HIT, EXP},			--139
	{HIT, MASTERY},		--140
	
	{CRIT, SPI},		--141
	{CRIT, DODGE},		--142
	{CRIT, PARRY},		--143
	{CRIT, HIT},		--144
	{CRIT, HASTE},		--145
	{CRIT, EXP},		--146
	{CRIT, MASTERY},	--147
	
	{HASTE, SPI},		--148
	{HASTE, DODGE},		--149
	{HASTE, PARRY},		--150
	{HASTE, HIT},		--151
	{HASTE, CRIT},		--152
	{HASTE, EXP},		--153
	{HASTE, MASTERY},	--154
	
	{EXP, SPI},			--155
	{EXP, DODGE},		--156
	{EXP, PARRY},		--157
	{EXP, HIT},			--158
	{EXP, CRIT},		--159
	{EXP, HASTE},		--160
	{EXP, MASTERY},		--161
	
	{MASTERY, SPI},		--162
	{MASTERY, DODGE},	--163
	{MASTERY, PARRY},	--164
	{MASTERY, HIT},		--165
	{MASTERY, CRIT},	--166
	{MASTERY, HASTE},	--167
	{MASTERY, EXP},		--168
}

---	Returns which stats that are modified by a given reforgeID.
-- The first return value is the id of the stat that was reduced.
-- The second return value is the id of the stat that was correspondingly increased.
-- If the passed reforgeID corresponded to an item that has not been reforged, then GetReforgedStatIDs returns nil, nil.
-- @param reforgeID Reforging ID to find the stat increase and decrease for, as returned by GetReforgeID
-- @return decreased stat number, increased stat number
--    returned stat id can be one of: (SPI, DODGE, PARRY, HIT, CRIT, HASTE, EXP, MASTERY)
--    If reforgeID corresponds to an item that wasn't reforged, the return values are nil
function lib:GetReforgedStatIDs(reforgeID)
	local rid = reforgeID - offset + 1
	local info = reforgeIDs[rid]
	if info then
		return info[1], info[2]
	else
		return nil, nil
	end
end

-- Returns information about the reforging with the given id.
-- Returns long stat names.
-- @param id Reforging ID as returned by GetReforgeID
-- @return decreased stat name, increased stat name
function lib:GetReforgedStatNames(reforgeID)
	local minus, plus = lib:GetReforgedStatIDs(reforgeID)
	if minus and plus then
		return lib.StatNames[minus], lib.StatNames[plus]
	else
		return nil, nil
	end
end

-- Returns information about the reforging with the given id.
-- Returns short stat names.
-- @param id Reforging ID as returned by GetReforgeID
-- @return decreased stat name, increased stat name
function lib:GetReforgedStatShortNames(id)
	local minus, plus = lib:GetReforgedStatIDs(id)
	if minus and plus then
		return lib.ShortStatNames[minus], lib.ShortStatNames[plus]
	else
		return nil, nil
	end
end

--- Extracts the Reforging ID from the item string or item link.
-- @param (itemString) An item string or item link
-- @return (number) The reforging ID applied to the item. If the item is not reforged the return value is nil.
function lib:GetReforgeID(itemString)
	--[[
		Item strings are of the format:
		
			item:[enchantId]:[jewelId1]:[jewelId2]:[jewelId3]:[jewelId4]:[suffixId]:[uniqueId]:[linkLevel]:[reforgeId]
		
		e.g.
			"item:12345:0:0:0:0:0:0:0"
			
		itemstrings for an item can be fetched using:
		
		
	--]]
	local id = tonumber(itemString:match("item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+:%d+:(%d+)"))
	return (id ~= 0) and id or nil
end

-- Checks if the given item is reforged
-- @param itemString An item string or item link
-- @return true if the item is reforged, false otherwise
function lib:IsItemReforged(itemString)
	local id = lib:GetReforgeID(itemString)
	return id ~= nil
end
