-- **************************************************************************
-- * MeleePoints.lua
-- * MeleePoints does stuff which i'm too lazy to describe here
-- **************************************************************************

local addonName = "MeleePoints";
local L = LibStub("AceLocale-3.0"):GetLocale(addonName) --Get the localization for our addon
local StatLogic = LibStub("LibStatLogic-1.2")
local TipHooker = LibStub("LibTipHooker-1.1")

MeleePoints = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0", "AceDebug-3.0", "AceTimer-3.0")
local MeleePoints = MeleePoints;

MeleePoints.version = GetAddOnMetadata(addonName, "Version");
MeleePoints.version = MeleePoints.version.." (r"..gsub("$Revision: 114 $", "$Revision: (%d+) %$", "%1")
--Append "a" (alpha) to revision if it's alpha. The wowace packager will convert alpha..end-alpha into a block level comment
--@alpha@
MeleePoints.version = MeleePoints.version.."a"
--@end-alpha@
MeleePoints.version = MeleePoints.version..")"
MeleePoints.date = gsub("$Date: 2010-12-14 13:30:31 +0000 (Tue, 14 Dec 2010) $", "^.-(%d%d%d%d%-%d%d%-%d%d).-$", "%1")

--MeleePoints.debugging = 3; 

MeleePoints.sourceData = {}; --unmodifed current player data
--MeleePoints.finalData = {}; 
--[[
	characterLevel = 85,
	mobLevelOffset = 3,
	weaponDamageMinBase = 660,
	weaponDamageMaxBase = 1226,
	weaponSpeedBase = 2.60, (seconds)
	
	attackSpeed = 2.5204, -- (seconds)						=weaponSpeedBase*(1-haste*0.01)
		haste = 3.061133026562, -- (%) 						=GetCombatRating(CR_HASTE_MELEE)
			hasteRating = 392, --							=GetCombatRatingBonus(CR_HASTE_MELEE)
	
	increasedDamagePerSwing = 975, --						=increasedDpsThroughAttackPower/weaponSpeedBase
		increasedDpsThroughAttackPower = 375.0714, --		=attackPower/14
			attackPower = 5251, 							=baseAttackPower+attackPowerThroughStrength
				baseAttackPower = 255,						=f(class,characterLevel)
				attackPowerThroughStrength = 4996,			=GetAttackPowerForStat(1, strength); --(1=Str)
					strength = 2508, 						=baseStrength+strengthFromItems
						baseStrength = 172,	
						strengthFromItems = 2336,
	
	weaponDamageMinWithAttackPower = 1635,					=weaponDamageMinBase+increasedDamagePerSwing
	weaponDamageMaxWithAttackPower = 2201,					=weaponDamageMaxBase+increasedDamagePerSwing
	
	modDamage = 2.0000, --(%)								=UnitDamage("player"), 7
	weaponDamageMin = 1668,									=weaponDamageMinWithAttackPower*(1+modDamage/100) =UnitDamage("player"), 1
	weaponDamageMax = 2245,									=weaponDamageMaxWithAttackPower*(1+modDamage/100) =UnitDamage("player"), 2
	
	weaponDamageAvg = 1957,									=(weaponDamageMin+weaponDamageMax)*0.5
	
	whiteDps = 776.3,										=weaponDamageAvg / attackSpeed
	
	---------------------------------------
	missChance = 5.7354, (%)								=GetMeleeMissChance(+3, false)+GetHitModifier()
	=GetMeleeMissChance(+3, false)+GetHitModifier()
	
	
	meleePoints = 3881.5,									=whiteDps / 0.20 (melee dps is, on average, 20% of melee damage. And since meleePoints is arbitrary scale: who cares)
]]--

---------------------
-- Local Variables --
---------------------
--Lua's compiler is not very good at calling functions, so we redeclare the functions here to help it call them faster ><
local GetMeleeMissChance = GetMeleeMissChance;			--PaperDollFrame\GetMeleeMissChance   Used by CalculateCombatTable
local GetEnemyDodgeChance = GetEnemyDodgeChance;		--PaperDollFrame\GetEnemyDodgeChance  Used by CalculateCombatTable
local GetEnemyDodgeChance = GetEnemyParryChance; 		--PaperDollFrame\GetEnemyParryChance  Used by CalculateCombatTable
local max = math.max;
local profileDB -- Initialized in :OnInitialize()

--[[
	When showing how much MeleePoints you get per unit of stat, this decides what the "per-unit" is.
	Normally you'd want it to be per 1 stat point:
		1 Strength = 2.728 MP
		1 Attack Power = 0.438 MP
		1 Haste Rating = 0.0178 MP
		1 Haste = 1.883 MP
	
	But maybe you want to see it in different scales:
		1 Strength = 2.782 MP
		2 Attack Power = 2.782 MP
		100 Haste Rating = 1.789 MP
		1.5 Haste = 3.281 MP
	This names of the entries in this table matche what you can pass in as changes to ApplyChanges(...)
--]]
local MeleePointsPerUnitSizes = {
		strength = 1.00,
		attackPower = 1.00,
		hasteRating = 1.00,
		hitRating = 1.00,
		expertiseRating = 1.00,
		agility = 1.00,
		critRating = 1.00,
}	

--- Bosses have a chance to be struck with a "glancing blow". These percentages are highly suspect
local BASE_GLANCING_CHANCE = {
	[0] =  6.17; --44/521(8.46%), 145/2315(6.26%), 379/6237(6.08%, 1.4s), 25/528(4.73%), 77/1351(5.70%, 1.4s), 55/897(6.13%, 3.8s), 91/1343(6.78%,1.4s)
	[1] = 14.00; --assume linear interpolation
	[2] = 19.50; --assume linear interpolation
	[3] = 24.00; --1898/8010(23.70%)  [http://www.mmo-champion.com/threads/775527-The-Cat-Guide-4.0.1-edition says 24%]
};	
--[[
	For level 85 target dummy
	For the raiders level target dummy (level skull, +3, 88), i ran three trials to find glancing blow chance:
		sample1: 1227/5175 23.71% (22.72% - 24.68% @90% confidence)
		sample2:  438/1877 23.34% (21.68% - 24.93% @90% confidence)
		sample3:  233/ 958 24.32% (21.92% - 26.62% @90% confidence)
		  total: 1898/8010 23.70% (22.91% - 22.48% @90% confidence)
		                          (22.76% - 24.73% @95% confidence)
--]]

--[[
	Miss Chance
	===============
	Level 88 raiders target dummy (level skull, +3, 88)
		sample1:  141/1668  8.453%
--]]


local BASE_ENEMY_BLOCK_CHANCE = {
	[0] = 4.80; --i think i read somewhere a base chance of 4.80%
	[1] = 4.95; --assume linear interpolation between +0 and +3
	[2] = 5.10; --assume linear interpolation between +0 and +3
	[3] = 5.24; --experimentally observed against Raiders Target Dummy. 5175 swings @271 blocked = 5.24% (20110110)
};

local BASE_ENEMY_CRIT_SUPPRESSION = {
	[0] = 0.00,
	[1] = 1.60,
	[2] = 3.20,
	[3] = 4.80,
}


GLANCING_BLOW_MODIFIER = 0.75; --glancing blows do (on average!) 25% less damage, i.e. 0.75
ENEMY_BLOCK_MODIFIER = 0.70;   --blocked attacks to 30% less damage, i.e. 0.70
--FRACTION_OF_TIME_IN_FRONT_OF_ENEMY = 0.20; --behind the mob 80% of the time, not as good at 100%, but better than 0%
-----------
-- Tools --
-----------
-- clear "to", and copy "from"
function copyTable(to, from)
	if to then
		for k in pairs(to) do
			to[k] = nil
		end
		setmetatable(to, nil)
	else
		to = {}
	end
	for k,v in pairs(from) do
		if type(k) == "table" then
			k = copyTable({}, k)
		end
		if type(v) == "table" then
			v = copyTable({}, v)
		end
		to[k] = v
	end
	setmetatable(to, getmetatable(from))
	return to
end

function table.val_to_str ( v )
  if "string" == type( v ) then
    v = string.gsub( v, "\n", "\\n" )
    if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  else
    return "table" == type( v ) and table.tostring( v ) or
      tostring( v )
  end
end

function table.key_to_str ( k )
  if "string" == type( k ) and string.match( k, "^[_%a][_%a%d]*$" ) then
    return k
  else
    return "[" .. table.val_to_str( k ) .. "]"
  end
end

function table.tostring( tbl )
  local result, done = {}, {}
  for k, v in ipairs( tbl ) do
    table.insert( result, table.val_to_str( v ) )
    done[ k ] = true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
      table.insert( result,
        table.key_to_str( k ) .. "=" .. table.val_to_str( v ) )
    end
  end
  return "{" .. table.concat( result, ", " ) .. "}"
end

function GetEnemyBlockChance(levelOffset)
	if (levelOffset < 0 or levelOffset > 3) then
		return 0;
	end
	local chance = BASE_ENEMY_BLOCK_CHANCE[levelOffset];
	
	if (chance < 0) then
		chance = 0;
	elseif (chance > 100) then
		chance = 100;
	end
	return chance;
end

function GetEnemyCritSuppression(levelOffset)
	if (levelOffset < 0 or levelOffset > 3) then
		return 0;
	end
	local chance = BASE_ENEMY_CRIT_SUPPRESSION[levelOffset];
	
	if (chance < 0) then
		chance = 0;
	elseif (chance > 100) then
		chance = 100;
	end
	return chance;
end

local dbDefaults = {
	profile = {
		mobLevelDiff = 0, -- [0..3] +0 = same level, +3 is three levels above (i.e. boss level)
		timeSpentBehind = 100, --percentage of time you're attacking the mob from behind (100=100% = always attacking boss from behind = best, 0=0% = always attacking boss from the front = worst)
		showMeleePointsDiff = true, --whether or not to show the meleepoints diff value in item tooltips. 
		showMeleePointsDiffPercent = true, --whether or not to show the meleepoints diff percentage in item tooltips
	}
}

function MeleePoints:OnInitialize()
    -- Called when the addon is loaded
	
	self.db = LibStub("AceDB-3.0"):New("MeleePointsDB", dbDefaults)
	-- Initialize profileDB
	profileDB = self.db.profile;

--	self.Print("SAuctionSold = "..SAuctionSold);
--	self.Print("SAuctionSoldItemPattern = "..SAuctionSold);
--	self.Print("SoundFileToPlay = "..soundFileToPlay);

	--Call SetupOptions if we've included the options file. (Not like there's any reason not to include it, its not like you can use the addon. But it's modular, helps with testing)
	if (self.SetupOptions) then
		self:SetupOptions() --in MeleePoints_Options.lua
	end
end

function MeleePoints:OnEnable()
    -- Called when the addon is enabled
	--self:RegisterEvent("CHAT_MSG_SYSTEM")
	self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "StatsChanged");	--Fires when the player equips or unequips an item
	self:RegisterEvent("UNIT_AURA", "StatsChanged");				--Fires when a unit loses or gains a buff or debuff.
	self:RegisterEvent("PLAYER_LEVEL_UP", "StatsChanged");			--Fires when the player gains a character level


	self:RegisterEvent("UNIT_AURA", "StatsChanged"); 
	self:RegisterEvent("UNIT_LEVEL", "StatsChanged");
	self:RegisterEvent("UNIT_STATS", "StatsChanged");
	self:RegisterEvent("UNIT_DAMAGE", "StatsChanged");				--Fires when a unit's weapon damage changes.
	self:RegisterEvent("UNIT_ATTACK_SPEED", "StatsChanged");		--Fires when a unit's attack speed changes
	self:RegisterEvent("UNIT_RANGEDDAMAGE", "StatsChanged");		--Fires when a unit's ranged attack damage changes
	self:RegisterEvent("UNIT_ATTACK", "StatsChanged");				--Fires when a unit's weapon (or standard melee attack damage) changes
	self:RegisterEvent("UNIT_STATS", "StatsChanged"); 				--Fires when a unit's primary attributes change. Primary attributes are Strength, Stamina, Agility, Intellect, and Spirit.
	self:RegisterEvent("UNIT_RANGED_ATTACK_POWER", "StatsChanged");	--Fires when a unit's ranged attack power changes

	--for the player 
	self:RegisterEvent("COMBAT_RATING_UPDATE", "StatsChanged");		--Fires when the player's combat rating statistics change
	self:RegisterEvent("MASTERY_UPDATE", "StatsChanged");			--Fires when the player's mastery rating is updated
	self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "StatsChanged");	--Fires when the player equips or unequips an item
	self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "StatsChanged");	--Fires when an effect changes the player's spell bonus damage

	--for other things
	self:RegisterEvent("PLAYER_TALENT_UPDATE", "StatsChanged");		--Fires when the player gains or spends talent points
	
	self:RegisterEvent("UNIT_ATTACK_POWER", "StatsChanged");		--Fires when a unit's attack power changes

	self:UpdateMeleePoints();

	-- Hook item tooltips
	if MeleePointsTooltips then
		MeleePointsTooltips:OnEnable();
	end;

end

function MeleePoints:OnDisable()
    -- Called when the addon is disabled
	self:CancelAllTimers();
	
	-- Unhook item tooltips
	if MeleePointsTooltips then
		MeleePointsTooltips:Disable();
	end;
end

function MeleePoints:StatsChanged(event, message)
	self:Debug(event);
	self:UpdateMeleePoints();
end

---Something global has changed, and meleepoints need to be updated, 
-- whether they are in the calculator, character sheet, or tooltip.
local tmrUpdate;
function MeleePoints:UpdateMeleePoints(sensder)
	MeleePoints:ClearCache(sender);

	--if (tmrUpdate) then
--		self:CancelTimer(tmrUpdate, true);
--	end;

	self:Debug("UpdateMeleePoints() - setting timer");
	if (tmrUpdate == nil) then
		tmrUpdate = self:ScheduleTimer("EndUpdateMeleePoints", 0.100, sender); --50ms = 20fps.  60fps = 16.666ms
	end;
end;

--- Update the sourceData member with the player's current stats
-- Call this method when attributes/buffs/equipped items for the player change.
-- This method will repopulate the sourceData member variable through a call to GetSourceData(...)
function MeleePoints:EndUpdateMeleePoints()
	self:Debug("MeleePoints:EndUpdateMeleePoints()");
	tmrUpdate = nil;
	
	--Invalidate any connected things (tooltip caches, lib data brokers, tital plugins, etc)
	if (self.UpdateLdb) then
		self:Debug("invalidting");
		self:UpdateLdb();
	end;
end

function MeleePoints:ClearCache(sender)
	MeleePoints.sourceData = nil;
end;

--- Fetch all the player's current stats, and store them in data
-- This function fetches the players current attributes and calcualtes their MeleePoints
function MeleePoints:GetSourceData(data)
	assert(data);
	
	local unit = "player"
	data.playerLevel = UnitLevel(unit)
	data.playerHealth = UnitHealthMax(unit)
	data.playerClass = select(2, UnitClass("player"))
	data.mobLevelDiff = (self.db.profile.mobLevelDiff or 0);
	data.mobLevel = UnitLevel(unit)+data.mobLevelDiff; -- + self.db.profile.mobLevelDiff
	data.timeSpentBehind = (self.db.profile.timeSpentBehind or 90); --percentage

	--[[
		The server only gives us weapon damage *after* power and damageModifier(%) have been applied.
		We have to back-calculate the original weapon damage values
	--]]
	local speed, _ = UnitAttackSpeed(unit);
	data.attackSpeed = speed;								--e.g. 2.5204 (seconds)
	local weaponDamageMin, weaponDamageMax, modDamage;
	weaponDamageMin, weaponDamageMax, _, _, _, _, modDamage = UnitDamage(unit); 

	data.modDamage = modDamage; 							--e.g. 2.0000; (%)
	data.weaponDamageMin = weaponDamageMin; 				--=weaponDamageMinWithAttackPower*(1+modDamage/100)  --e.g. 1668
	data.weaponDamageMax = weaponDamageMax; 				--=weaponDamageMaxWithAttackPower*(1+modDamage/100)  --e.g. 2245
	data.weaponDamageAvg = (data.weaponDamageMin+data.weaponDamageMax)*0.5; 	
	data.whiteDps = data.weaponDamageAvg / data.attackSpeed; --776.3

	--[[
		We now reconstruct all base everthing leading up to that, so that we can play what-if scenarios
		Base weapon values have to be back-calculated (server only tells us numbers after power and DamageMod applied, and after haste)
	]]--
	--Get original weapon speed (without haste applied)
	data.hasteRating = GetCombatRating(CR_HASTE_MELEE); 	--392;
	data.haste = GetCombatRatingBonus(CR_HASTE_MELEE); 		--3.061133026562; --%    2.6*(1-0.0306113302) = 2.5204
	data.weaponSpeedBase = data.attackSpeed / (1-data.haste*0.01); 	--2.6000 = 2.5204/(1-0.030611302) (seconds)

	--Remove the damageMod to get "weapon damage with attack power". 
	--damage = damageWithAttackPower*(1+modDamage/100)
	data.weaponDamageMinWithAttackPower = data.weaponDamageMin / (1+data.modDamage*0.01) --1635; =weaponDamageMinBase+increasedDamagePerSwing
	data.weaponDamageMaxWithAttackPower = data.weaponDamageMax / (1+data.modDamage*0.01) --2201; =weaponDamageMaxBase+increasedDamagePerSwing

	
	--AttackPower <- BaseAttackPower, AttackPowerThroughStrength <- Strength
	local stat, effectiveStat, posBuff, negBuff = UnitStat(unit, 1); --1=Strength
	--stat, effectiveStat, posBuff, negBuff = UnitStat(unit, statIndex);
	data.strength = stat; --								=baseStrength+strengthFromItems  --2508
	data.attackPowerThroughStrength = GetAttackPowerForStat(1, data.strength); --(1=Str)   --4996

	local base, posBuff, negBuff = UnitAttackPower(unit);
	data.attackPower = base; 								--5251; =baseAttackPower+attackPowerThroughStrength
	data.increasedDpsThroughAttackPower = data.attackPower/14;	--375.0714    =attackPower/14
	data.increasedDamagePerSwing = data.increasedDpsThroughAttackPower*data.weaponSpeedBase; --975; =increasedDamageThroughAttackPower*weaponSpeedBase
	--data.baseStrength = 172;	
	--data.strengthFromItems = 2336;

	data.baseAttackPower = data.attackPower - data.attackPowerThroughStrength; -- 255; --							=f(class,characterLevel)
	
	data.weaponDamageMinBase = data.weaponDamageMinWithAttackPower - data.increasedDamagePerSwing; --660;
	data.weaponDamageMaxBase = data.weaponDamageMaxWithAttackPower - data.increasedDamagePerSwing; --1226;
	
	--Now stats needed by the combat table, (agility, crit rating, crit, hit, hit rating, expertise rating, expertise)
	data.hitRating = GetCombatRating(CR_HIT_MELEE); --e.g. 208 hit rating
	data.hit = GetCombatRatingBonus(CR_HIT_MELEE); --e.g. 1.7317631610591%
	
	
	--[[
		Expertise (main, offHand) = GetExpertise();
		Expertise(%) (main, offHand) = GetExpertisePercent()   -- lowers chance to be dodged/parried by xx%, yy%
		
		ExpertiseRating = GetCombatRating(CR_EXPERTISE)
		ExpertiseFromExpertiseRating(%) = GetCombatRatingBonus(CR_EXPERTISE)

		Players have some in-build expertise.
		ExpertisePercent = f(GetExpertise)
		                 = f(Expertise + [baseExpertiseRating])
		
	--]]
	data.expertiseRating = GetCombatRating(CR_EXPERTISE); --e.g. 324 Expertise Rating

	local expertise, expertiseFromRating = GetExpertise(); --e.g. 13, 10. Player has 13 Expertise (+10 Expertise from Expertise Rating)
	data.expertise = expertise; --expertisePercent = expertise / 4 (0.25% per point of Expertise)
	data.expertiseBase = expertise-expertiseFromRating; --expertise = expertiseBase + expertiseFromExpertiseRating

	local expertisePercent, expertisePercentFromRating = GetExpertisePercent();	 --e.g. 3.25, 2.5.
	data.expertisePercent = expertisePercent;
	
	--[[
		Crit Rating
		Crit
	--]]
	data.agility = UnitStat(unit, 2);  --2 = agility
	data.critRating = GetCombatRating(CR_CRIT_MELEE);
	data.crit = GetCritChance();


	--Combat Table
	self:CalculateCombatTable(data, data.timeSpentBehind*0.01);
end;

--[[
This function returns current melee points
--]]
function MeleePoints:GetCurrentMeleePoints()

	if (self.sourceData == nil) then
		self.sourceData = {};
	end

	self:GetSourceData(self.sourceData)
	
	--self:Debug(string.format("MeleePoints=%.1f", self.sourceData.meleePoints));

	return MeleePoints.sourceData;
end;

--- Calculate the combat table based on values inside data and store it in data.
-- @param data A MeleePoints table
-- @param fractionOfTimeBehindEnemy 1=always behind enemy (best; can never be parried or blocked), 0=never behind enemy (worst; attacks can be blocked/parried)
function MeleePoints:CalculateCombatTable(data, fractionOfTimeBehindEnemy)
	assert(data)
	assert(fractionOfTimeBehindEnemy);
	assert(fractionOfTimeBehindEnemy >= 0);
	assert(fractionOfTimeBehindEnemy <= 1);
	
	assert(data.mobLevelDiff >= 0 and data.mobLevelDiff <= 3)

	local fractionOfTimeInFrontOfEnemy = (1-fractionOfTimeBehindEnemy);
	local chanceRemaining = 100.0;
	
	--Miss Chance <-- Hit (%) <-- Hit Rating
	data.missChance = min(max(BASE_MISS_CHANCE_PHYSICAL[data.mobLevelDiff]-data.hit, 0), chanceRemaining); ---+GetHitModifier(); 	--5.7354 (%) =PaperDollFrame\GetMeleeMissChance
	chanceRemaining = max(chanceRemaining - data.missChance, 0);
	
	--Enemy Dodge chance <-- Expertise(%) <-- Expertise Rating
	local dodgeChance = max(BASE_ENEMY_DODGE_CHANCE[data.mobLevelDiff]-data.expertisePercent, 0);
	data.enemyDodgeChance = min(dodgeChance, chanceRemaining);
	chanceRemaining = max(chanceRemaining-data.enemyDodgeChance, 0);

	--Enemy Parry chance <-- Expertise(%) <-- Expertise Rating (and standing behind him)
	local parryChance = max(BASE_ENEMY_PARRY_CHANCE[data.mobLevelDiff]-data.expertisePercent, 0) * (fractionOfTimeInFrontOfEnemy);
	data.enemyParryChance = min(parryChance, chanceRemaining); --chance, offhandChance = PaperDollFrame\GetEnemyParryChance(levelOffset)
	chanceRemaining = max(chanceRemaining-data.enemyParryChance, 0);

	--There doesn't seem to be anything that can affect the chance of glancing blows, except the level differential
	data.glancingChance = min(BASE_GLANCING_CHANCE[data.mobLevelDiff], chanceRemaining);
	chanceRemaining = max(chanceRemaining-data.glancingChance, 0);

	--There's nothing that can affect an enemy blocking, except standing behind him
	local blockChance = BASE_ENEMY_BLOCK_CHANCE[data.mobLevelDiff] * (fractionOfTimeInFrontOfEnemy);
	data.blockChance = min(blockChance, chanceRemaining);
	chanceRemaining = max(chanceRemaining-data.blockChance, 0);
	
--	self:Print(string.format("chanceRemaining: %.2f", chanceRemaining));
--	self:Print(string.format("crit suppression %.2f", GetEnemyCritSuppression(data.mobLevelDiff)));
	
	local critChance = max(data.crit - GetEnemyCritSuppression(data.mobLevelDiff), 0);
	data.critChance = min(critChance, chanceRemaining);
	--self:Print(string.format("critChange: %.2f%%", data.critChance));
	chanceRemaining = max(chanceRemaining-data.critChance, 0);
	
	data.hitChance = chanceRemaining;
	
	--mob armor reduces their damage taken
	local mobArmor = 10000; --i'm getting 30.04% reduction on a mix of 84-85 Twilight Highland mobs, which works out to 10,000-11,200 armor
	data.damageReductionFromArmor = StatLogic:GetReductionFromArmor(mobArmor, data.playerLevel); --return reduction fraction, e.g. 0.23489 ==> 23.489% reduction
	
	--add it all up into a net factor
	data.combatTableDamageMultiplier = 
			--0*data.missChance + 		--missed attacks do zero damage
			--0*data.dodgeChance +		--dodged attacks do zero damage
			--0*data.parryChance +		--parried attacks do zero damage
			GLANCING_BLOW_MODIFIER*data.glancingChance +	--glancing blows do (on average!) 25% less damage, i.e. 0.75
			ENEMY_BLOCK_MODIFIER*data.blockChance + 		--blocked attacks to 30% less damage, i.e. 0.70
			2*data.critChance +
			data.hitChance;

	data.combatTableDamageMultiplier = data.combatTableDamageMultiplier*(1-data.damageReductionFromArmor);
			
	
	data.whiteDpsAfterCombatTable = data.whiteDps * data.combatTableDamageMultiplier*0.01;
	data.meleePoints = data.whiteDpsAfterCombatTable / 0.20;
end

function MeleePoints:Test()
	local target = {};
	
	assert(self.sourceData);
	copyTable(target, self.sourceData);
	assert(target);
	
	local changes = {
		hasteRating = 1,
		haste = 1,
		strength = 1,
		attackPower = 1,
		hitRating = 1,
		hit = 1,
		expertise = 1,
		expertisePercent = 1,
		agility = 1,
		critRating = 1,
		crit = 1,
	}
	
	self:ApplyChanges(target, changes);
	
	--self:ProcessSystemMessage(testMessage);
end;

function MeleePoints:ProcessSystemMessage(message)
	self:Debug(string.format("ProcessSystemMessage(%s)", message));
end

function MeleePoints:DumpTable()
	local data = self.sourceData;
	
	self:DumpDataTable(data);
end

function MeleePoints:DumpDataTable(data)
	
	assert(data);

	self:Print("MeleePoints.sourceData");
	self:Print(string.format("   playerLevel: %d", data.playerLevel));
	self:Print(string.format("   playerHealth: %d", data.playerHealth));
	self:Print(string.format("   playerClass: %s", data.playerClass));
	self:Print(string.format("   mobLevel: %d", data.mobLevel));
	self:Print(string.format("   mobLevelDiff: +%d", data.mobLevelDiff));
	self:Print(string.format("   timeSpentBehind: %d%%", data.timeSpentBehind));

	--white damage attributes
	self:Print(string.format("   weaponDamageMinBase: %d", data.weaponDamageMinBase));
	self:Print(string.format("   weaponDamageMaxBase: %d", data.weaponDamageMaxBase));
	self:Print(string.format("   weaponSpeedBase: %.2f", data.weaponSpeedBase));
	
	self:Print(string.format("   hasteRating: %d", data.hasteRating));
	self:Print(string.format("   haste: %.4f%%", data.haste));
	self:Print(string.format("   attackSpeed: %.4f", data.attackSpeed));
	
	self:Print(string.format("   strength: %d", data.strength));
	self:Print(string.format("   attackPowerThroughStrength: %d", data.attackPowerThroughStrength));
	self:Print(string.format("   baseAttackPower: %d", data.baseAttackPower));
	self:Print(string.format("   attackPower: %d", data.attackPower));
	
	self:Print(string.format("   increasedDamagePerSwing: %d", data.increasedDamagePerSwing));
	self:Print(string.format("   increasedDpsThroughAttackPower: %.4f", data.increasedDpsThroughAttackPower));
	--self:Print(string.format("   baseStrength: %d", data.baseStrength));
	--self:Print(string.format("   strengthFromItems: %d", data.strengthFromItems));
	
	self:Print(string.format("   weaponDamageMinWithAttackPower: %d", data.weaponDamageMinWithAttackPower));
	self:Print(string.format("   weaponDamageMaxWithAttackPower: %d", data.weaponDamageMaxWithAttackPower));
	
	self:Print(string.format("   modDamage: %.4f%%", data.modDamage));
	self:Print(string.format("   weaponDamageMin: %d", data.weaponDamageMin));
	self:Print(string.format("   weaponDamageMax: %d", data.weaponDamageMax));
	
	self:Print(string.format("   weaponDamageAvg: %.1f", data.weaponDamageAvg));
	
	self:Print(string.format("   whiteDps: %.2f", data.whiteDps));
	
	self:Print(string.format("   hitRating: %d", data.hitRating));
	self:Print(string.format("   hit: %.4f%%", data.hit));
	
	self:Print(string.format("   expertise: %d", data.expertise));
	self:Print(string.format("   expertisePercent: %.4f%%", data.expertisePercent));
	
	self:Print(string.format("   critRating: %d", data.critRating));
	self:Print(string.format("   crit: %.4%%f", data.crit));

	
	self:Print(string.format("   missChance: %.2f%%", data.missChance));
	self:Print(string.format("   enemydodgeChance: %.2f%%", data.enemyDodgeChance));
	self:Print(string.format("   enemyparryChance: %.2f%%", data.enemyParryChance));
	self:Print(string.format("   glancingChance: %.2f%%", data.glancingChance));
	self:Print(string.format("   blockChance: %.2f%%", data.blockChance));
	self:Print(string.format("   critChance: %.2f%%", data.critChance));
	self:Print(string.format("   hitChance: %.2f%%", data.hitChance));
	self:Print(string.format("   combatTableDamageMultiplier: %.2f%%", data.combatTableDamageMultiplier));
	
	self:Print(string.format("   meleePoints: %.1f", data.meleePoints));
end;

function MeleePoints:ApplyChanges(target, changes)

	assert(target);
	assert(changes);
	
	local function LogChange(s, ...)
		self:Debug(string.format(s, ...));
	end;

	if (changes.hasteRating and changes.hasteRating ~= 0) then
		--[[
			The WowApi GetCombatRatingBonus(CR_HASTE_MELEE) tells you your Haste% bonus based on your current HasteRating.
			We need to use RatingBuster's version to calculate the same thing for an arbitrary haste rating
			
			Haste looks to be linearly related HasteRating.
		--]]
		local hasteFromHasteRating;
		hasteFromHasteRating = StatLogic:GetEffectFromRating(changes.hasteRating, CR_HASTE_MELEE)
		
		LogChange("Adding %.4f%% Haste (from %d Haste Rating) to existing %.4f%% Haste", hasteFromHasteRating, changes.hasteRating, target.haste);
		
		target.hasteRating = target.hasteRating + changes.hasteRating;
		
		changes.haste = (changes.haste or 0) + hasteFromHasteRating;
	end;
	
	if (changes.haste and changes.haste ~= 0) then
		--[[
			Haste is a percentage, that affects base weapon speed.
		--]]
		target.haste = target.haste + changes.haste;
		
		LogChange("Appling %.4f%% Haste to exising weapon speed of %.4f seconds", changes.haste, target.weaponSpeedBase);
		target.attackSpeed = target.weaponSpeedBase*(1-target.haste*0.01)
		LogChange("   New Attack Speed: %.4f", target.attackSpeed);
	end;

	if (changes.strength and changes.strength ~= 0) then

		--[[
			If we have a modifier in effect on us, then apply that same modifier to the new strength amount
		--]]
		local modStr = StatLogic:GetStatMod("MOD_STR"); --e.g. 1.1025 = 110.25% = increase of 10.25%
		if (modStr ~= 1.0) then
			LogChange("Altering %.4f Strength change by MOD_STR %.4f", changes.strength, modStr);
			changes.strength = changes.strength * modStr;
		end
		
		--Add strength to target
		target.strength = (target.strength or 0) + changes.strength;
		LogChange("   New Strength: %.2f", target.strength);
		
		
		--[[
			Now bump up AttackPower<-Strength
			The WoWApi GetAttackPowerForStat(1, strength) returns the attack power provided by strength, 
			but it's a non-linear scale; fortunatly it *is* linear at the top.
		--]]
		local attackPowerThroughStrength = StatLogic:GetAPFromStr(changes.strength);

		LogChange("Adding %.4f Attack Power (from %d Strength) to existing Attack Power of %d", attackPowerThroughStrength, changes.strength, target.attackPower);
		changes.attackPower = (changes.attackPower or 0) + attackPowerThroughStrength;
	end;

	if (changes.attackPower and changes.attackPower ~= 0) then
		--[[
			Attack Power grants DPS, which is damage per base weapon speed
		--]]
		local modAP = StatLogic:GetStatMod("MOD_AP"); --e.g.
		if (modAP ~= 1.0) then
			LogChange("Altering %.4f Attack Power change by MOD_AP %.4f", changes.attackPower, modAP);
			changes.attackPower = changes.attackPower * modAP;
		end
		
		--Add AttackPower to target
		target.attackPower = target.attackPower + changes.attackPower
		LogChange("   New Attack Power: %.2f", target.attackPower);
		
		--Now update weapon damage increase values in target because of change in attack power
		local increasedDpsThroughAttackPower = (changes.attackPower / 14);

		LogChange("Adding %.4f DPS (through %d Attack Power) to existing %.4f DPS Through Attack Power", increasedDpsThroughAttackPower, changes.attackPower, target.increasedDpsThroughAttackPower);
		
		target.increasedDpsThroughAttackPower = target.increasedDpsThroughAttackPower + increasedDpsThroughAttackPower;
		LogChange("   New dps through attack power: %.2f", target.increasedDpsThroughAttackPower);
		
		--Update the increasedDamagePerSwing value, since AttackPower->DPS changed
		target.increasedDamagePerSwing = target.increasedDpsThroughAttackPower * target.weaponSpeedBase;
		LogChange("   New Increased damage per swing: %.2f", target.increasedDamagePerSwing);

		target.weaponDamageMinWithAttackPower = target.weaponDamageMinBase + target.increasedDamagePerSwing
		target.weaponDamageMaxWithAttackPower = target.weaponDamageMaxBase + target.increasedDamagePerSwing

		target.weaponDamageMin = target.weaponDamageMinWithAttackPower*(1+target.modDamage/100);
		target.weaponDamageMax = target.weaponDamageMaxWithAttackPower*(1+target.modDamage/100); 
	
		LogChange("   Final WeaponDamageMin (with modDamage): %.2f", target.weaponDamageMin);
		LogChange("   Final WeaponDamageMax (with modDamage): %.2f", target.weaponDamageMax);
	
		target.weaponDamageAvg = (target.weaponDamageMin+target.weaponDamageMax)*0.5;
	end;
	
	target.whiteDps = target.weaponDamageAvg / target.attackSpeed;
	self:Debug(string.format("ApplyChanges(...): New whiteDps: %.2f", target.whiteDps));
	
	------------------------------------------
	-- Changes that affect the combat table --
	------------------------------------------
	--Hit<-HitRating
	if (changes.hitRating) and (changes.hitRating ~= 0) then
		LogChange("Adding %.2f Hit Rating to existing Hit Rating of %.2f", changes.hitRating, target.hitRating);
		target.hitRating = max(target.hitRating + changes.hitRating, 0);
		
		-- Hit<-HitRating
		local hitFromHitRating = StatLogic:GetEffectFromRating(changes.hitRating, CR_HIT_MELEE);
		LogChange("Converting %.2f Hit Rating to %.4f Hit", changes.hitRating, hitFromHitRating);
		changes.hit = (changes.hit or 0) + hitFromHitRating;
	end
	
	--Hit
	if (changes.hit) and (changes.hit ~= 0) then
		LogChange("Adding %.4f%% Hit to existing Hit of %.4f%%", changes.hit, target.hit);
		target.hit = max(target.hit + changes.hit, 0);
	end
	
	--- Expertise<-ExpertiseRating
	if (changes.expertiseRating and changes.expertiseRating ~= 0) then
		LogChange("Adding %d Expertise Rating to existing Expertise Rating of %d", changes.expertiseRating, target.expertiseRating);
		target.expertiseRating = max(target.expertiseRating + changes.expertiseRating, 0);
		
		LogChange("   New Expertise Rating: %d", target.expertiseRating);
		
		local expertiseFromExpertiseRating = StatLogic:GetEffectFromRating(target.expertiseRating, CR_EXPERTISE)
		--expertiseFromExpertiseRating = floor(expertiseFromExpertiseRating); --expertise is floored, e.g. 56 Expertise Rating = 1.8649740837514 ==> 1
		LogChange("Converting %d Expertise Rating to Expertise of %d", target.expertiseRating, expertiseFromExpertiseRating);
		target.expertise = target.expertiseBase + expertiseFromExpertiseRating;
		LogChange("   New Expertise: %d", target.expertise);
		target.expertisePercent = target.expertise / 4; --1 Expertise = 0.25%

	end
	
	--- Expertise
	if (changes.expertise and changes.expertise ~= 0) then
		LogChange("Adding %d Expertise to existing Expertise of %d", changes.expertise, target.expertise)
		target.expertise = max(target.expertise + changes.expertise, target.expertiseBase);
		LogChange("   New Expertise: %d", target.expertise);
		target.expertisePercent = target.expertise / 4; --1 Expertise = 0.25%
	end
	
	--CrititicalChanceBase<-Agility
	if (changes.agility and changes.agility ~= 0) then
		local modAgi = StatLogic:GetStatMod("MOD_AGI"); --e.g. 1.1025 = 110.25% = increase of 10.25%
		if (modAgi ~= 1.0) then
			LogChange("Altering %.4f Agility change by MOD_AGI %.4f", changes.agility, modAgi);
			changes.agility = changes.agility * modAgi;
		end

	
		LogChange("Adding %.4f Agility to existing Agility of %.4f", changes.agility, target.agility);
		target.agility = max(target.agility + changes.agility, 0);
	
		-- Crit<-Agility
		local critFromAgility = StatLogic:GetCritFromAgi(changes.agility);
		LogChange("Converting %.4f Agility into %.4f%% Crit", changes.agility, critFromAgility);
		changes.crit = (changes.crit or 0) + critFromAgility;
	end;
	
	--CrititicalChanceBase<-CritRating
	if (changes.critRating and changes.critRating ~= 0) then
		LogChange("Adding %.4f Crit Rating to existing Crit Rating of %.4f", changes.critRating, target.critRating);
		target.critRating = max(target.critRating + changes.critRating, 0);
	
		--Crit rating increases crit
		local critFromCritRating = StatLogic:GetEffectFromRating(changes.critRating, CR_CRIT_MELEE);
		LogChange("Converting %.2f Crit Rating to Crit %.4f%%", changes.critRating, critFromCritRating);
		changes.crit = (changes.crit or 0) + critFromCritRating;
	end
	
	--CrititicalChanceBase
	if (changes.crit and changes.crit ~= 0) then
		LogChange("Adding %.4f%% Critical Strike Chance to existing Critical Strike Chance of %.4f%%", changes.crit, target.crit);
		target.crit = max(target.crit + changes.crit, 0);
	end
	
	
	
	-------------------------
	--Combat table entries --
	-------------------------
	self:CalculateCombatTable(target, target.timeSpentBehind*0.01);
	
	self:Debug(string.format("ApplyChanges(...): New meleePoints: %.2f", target.meleePoints));
end

function MeleePoints:DumpStatValues()
	assert(self.sourceData);

	local target = {};
	local changes = {};
	
	local baseMeleePoints = self.sourceData.meleePoints;
	local diff = 0;
	
	--Strength
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.strength;
	changes = { strength = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Strength", target.meleePoints-baseMeleePoints, "MP"))

	--Attack Power
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.attackPower;
	changes = { attackPower = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Attack Power", target.meleePoints-baseMeleePoints, "MP"))

	--Haste Rating
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.hasteRating;
	changes = { hasteRating = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Haste Rating", target.meleePoints-baseMeleePoints, "MP"))

	--Hit Rating
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.hitRating;
	changes = { hitRating = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Hit Rating", target.meleePoints-baseMeleePoints, "MP"))

	--Expertise
	--[[
		Expertise from Expertise Rating is floored, so it takes about 23 Expertise Rating before you gain 1 Expertise.
		When we perform the "per 1 change" calculation, we don't floor the resulting Expertise - so that we can get
		some idea of the value of 1 Expertise Rating.
		Problem with that is that your current Expertise is already using a floored Expertise Rating value.
		Suddenly we compare a floored Expertise Rating against a non-rounded Expertise Rating, we're going to
		instantly see a huge jump in melee points. 
		What we need to do is compare your "current" melee points, but without flooring Expertise Rating.
		Then we run it again adding our +1 Expertise Rating, and we'll see a more representative value.
		We'll do this my running the ApplyChanges twice, once with our desired +1, and once with +0.000001
		(the ApplyChanges routine filters out a +0 change, so we'll feed it an arbitrarily low number)
	--]]
	copyTable(target, self.sourceData);
	diff = 0.0000001;
	changes = { expertiseRating = diff };
	self:ApplyChanges(target, changes);
	local baseMeleePointsWithExpertiseNotFloored = target.meleePoints;
	
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.expertiseRating;
	changes = { expertiseRating = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Expertise Rating", target.meleePoints-baseMeleePointsWithExpertiseNotFloored, "MP"))

	--Agility
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.agility;
	changes = { agility = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Agility", target.meleePoints-baseMeleePoints, "MP"))

	--Crit Rating
	copyTable(target, self.sourceData);
	diff = MeleePointsPerUnitSizes.critRating;
	changes = { critRating = diff };
	self:ApplyChanges(target, changes);
	self:Print(string.format("   %.2f %s = %.3f %s", diff, "Crit Rating", target.meleePoints-baseMeleePoints, "MP"))
end;

function MeleePoints:GetRelativeStatValues(sourceData)
--[[
	Returns a table containing the relative values of the stats:
		{
			strength = 0.47,
			attackPower = 0.33,
			hasteRating = 0.338,
			hitRating = 0.338,
			expertiseRating = 0.37,
			agility = 0.38,
			critRating = 0.332,
		}
--]]
	assert(sourceData);

	local target = {};
	local changes = {};
	
	local baseMeleePoints = sourceData.meleePoints;
	local diff = 0;
	
	local result = {};
	
	--Strength
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.strength;
	changes = { strength = diff };
	self:ApplyChanges(target, changes);
	result.strength = target.meleePoints-baseMeleePoints;

	--Attack Power
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.attackPower;
	changes = { attackPower = diff };
	self:ApplyChanges(target, changes);
	result.attackPower = target.meleePoints-baseMeleePoints;

	--Haste Rating
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.hasteRating;
	changes = { hasteRating = diff };
	self:ApplyChanges(target, changes);
	result.hasteRating = target.meleePoints-baseMeleePoints;

	--Hit Rating
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.hitRating;
	changes = { hitRating = diff };
	self:ApplyChanges(target, changes);
	result.hitRating = target.meleePoints-baseMeleePoints;

	--Expertise
	--[[
		Expertise from Expertise Rating is floored, so it takes about 23 Expertise Rating before you gain 1 Expertise.
		When we perform the "per 1 change" calculation, we don't floor the resulting Expertise - so that we can get
		some idea of the value of 1 Expertise Rating.
		Problem with that is that your current Expertise is already using a floored Expertise Rating value.
		Suddenly we compare a floored Expertise Rating against a non-rounded Expertise Rating, we're going to
		instantly see a huge jump in melee points. 
		What we need to do is compare your "current" melee points, but without flooring Expertise Rating.
		Then we run it again adding our +1 Expertise Rating, and we'll see a more representative value.
		We'll do this my running the ApplyChanges twice, once with our desired +1, and once with +0.000001
		(the ApplyChanges routine filters out a +0 change, so we'll feed it an arbitrarily low number)
	--]]
	copyTable(target, sourceData);
	diff = 0.0000001;
	changes = { expertiseRating = diff };
	self:ApplyChanges(target, changes);
	local baseMeleePointsWithExpertiseNotFloored = target.meleePoints;
	
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.expertiseRating;
	changes = { expertiseRating = diff };
	self:ApplyChanges(target, changes);
	result.expertiseRating = target.meleePoints-baseMeleePointsWithExpertiseNotFloored;

	--Agility
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.agility;
	changes = { agility = diff };
	self:ApplyChanges(target, changes);
	result.agility = target.meleePoints-baseMeleePoints;

	--Crit Rating
	copyTable(target, sourceData);
	diff = MeleePointsPerUnitSizes.critRating;
	changes = { critRating = diff };
	self:ApplyChanges(target, changes);
	result.critRating = target.meleePoints-baseMeleePoints;
	
	return result;
end;



function table.val_to_str ( v )
  if "string" == type( v ) then
    v = string.gsub( v, "\n", "\\n" )
    if string.match( string.gsub(v,"[^'\"]",""), '^"+$' ) then
      return "'" .. v .. "'"
    end
    return '"' .. string.gsub(v,'"', '\\"' ) .. '"'
  else
    return "table" == type( v ) and table.tostring( v ) or
      tostring( v )
  end
end

function table.key_to_str ( k )
  if "string" == type( k ) and string.match( k, "^[_%a][_%a%d]*$" ) then
    return k
  else
    return "[" .. table.val_to_str( k ) .. "]"
  end
end

function table.tostring( tbl )
  local result, done = {}, {}
  for k, v in ipairs( tbl ) do
    table.insert( result, table.val_to_str( v ) )
    done[ k ] = true
  end
  for k, v in pairs( tbl ) do
    if not done[ k ] then
      table.insert( result,
        table.key_to_str( k ) .. "=" .. table.val_to_str( v ) )
    end
  end
  return "{" .. table.concat( result, ", " ) .. "}"
end

function MeleePoints:VarAsString(value)
--[[Convert a variable to a string
--]]

	local Result = "["..type(value).."] = "
	
	--[[
		Some types in LUA refuse to be converted to a string, so we have to do it for it.
	
		LUA type() function returns a lowercase string that contains one of the following:
			- "nil"			we must manually return "nil"
			- "boolean"		we must manually convert to "true" or "false"
			- "number"
			- "string"
			- "function"
			- "userdata"
			- "thread"
			- "table"		we must manually convert to a string
			
	]]--
	
	if (value == nil) then
		Result = Result.."nil"
	elseif (type(value) == "table") then
		Result = Result..table.tostring(value)
	elseif (type(value) == "boolean") then
		if (value) then
			Result = Result.."true"
		else
			Result = Result.."false"
		end
	else
		Result = Result..value
	end
	
	return Result;
end

---------------------------------------------------------
-- Toggle the MeleePoints calculator, if it's available --
---------------------------------------------------------
function MeleePoints:ToggleCalculator()
	local mpc = MeleePointsCalculatorFrame;
	if (mpc) then
		if(mpc:IsVisible()) then
			mpc:Hide()
		else
			mpc:Show()
		end
	end
end;