local addon = MeleePoints; 

local L = LibStub("AceLocale-3.0"):GetLocale("MeleePoints") --Get the localization for our addon
local ldb = LibStub:GetLibrary("LibDataBroker-1.1");
local Toolkit = LibStub("Toolkit-1.0", 2);

local dataObject;

if (ldb == nil) then
	--if we couldn't load LibDataBroker then don't bother continuing
	return;
end;

dataObject = ldb:NewDataObject("MeleePoints", {
		type = "data source", --"launcher",
		text = "This text is required but never used",
		value = "1,234",
		suffix = L["MP"], --abbreviation of "MeleePoints"
		label = "MeleePoints",
		icon = "Interface\\Icons\\Inv_sword_27",
		OnClick = function(clickedframe, button)
			if (button == "RightButton") then
				addon:ShowConfig(); --InterfaceOptionsFrame_OpenToCategory(addon.optionsFrames.general)					 
			else
				addon:ToggleCalculator();
			end
		end
});

--Thrash Blade: Inv sword 36
--Gray sword: Inv_sword_04
--Warrior Arms: Inv sword 27

function dataObject:OnTooltipShow()
	self:AddLine(HIGHLIGHT_FONT_COLOR_CODE..L["MeleePoints"]..FONT_COLOR_CODE_CLOSE); --|cffffcc00|r

	local data = addon:GetCurrentMeleePoints();

	addon:LevelDebug(3, Toolkit:VarAsString(data));

	assert(data, "MeleePoints dataObject:OnTooltipShow: addon:GetCurrentMeleePoints returned nil");

	self:AddDoubleLine("Paperdoll Dps:",				HIGHLIGHT_FONT_COLOR_CODE..format("%.1f",	data.whiteDps)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Combat Table Scaling (%):",		HIGHLIGHT_FONT_COLOR_CODE..format("%.1f%%",	data.combatTableDamageMultiplier)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("MeleePoints:",					HIGHLIGHT_FONT_COLOR_CODE..format("%.1f",	data.meleePoints)..FONT_COLOR_CODE_CLOSE);
	self:AddLine(" ");
	
	self:AddDoubleLine("Mob Level:",					HIGHLIGHT_FONT_COLOR_CODE..format("+%d (%d)", data.mobLevel-data.playerLevel, data.mobLevel)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Time attacking from behind:",	Toolkit:RangeColorText(format("%d%%", data.timeSpentBehind), data.timeSpentBehind, 0, 1, 0, 0, 100, 1, 1, 1));
		--TS.GetRangeColoredText(format("%d%%", data.timeSpentBehind), data.timeSpentBehind, 0, 1, 0, 0, 100, 1, 1, 1).."\n"..
	self:AddLine(" ");

	self:AddLine(HIGHLIGHT_FONT_COLOR_CODE.."Combat Table"..FONT_COLOR_CODE_CLOSE)
	self:AddDoubleLine("Miss Chance:   |cff8080800% dmg|r",				HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.missChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Enemy Dodge Chance:   |cff8080800% dmg|r",		HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.enemyDodgeChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Enemy Parry Chance:   |cff8080800% dmg|r",		HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.enemyParryChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Glancing Chance:   |cff80808080% dmg|r",		HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.glancingChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Block Chance:   |cff80808070% dmg|r",			HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.blockChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Crit Chance:   |cff808080200% dmg|r",			HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.critChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Hit Chance:   |cff808080100% dmg|r",			HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.hitChance)..FONT_COLOR_CODE_CLOSE);
	self:AddDoubleLine("Armor Reduction:",								HIGHLIGHT_FONT_COLOR_CODE..format("%.2f%%", data.damageReductionFromArmor*100)..FONT_COLOR_CODE_CLOSE);
			
	self:AddDoubleLine("Total:",										HIGHLIGHT_FONT_COLOR_CODE..format("%.1f%%",	data.combatTableDamageMultiplier)..FONT_COLOR_CODE_CLOSE);


	local relativeStatValues = addon:GetRelativeStatValues(data);
	
	local minRel = math.min(
			relativeStatValues.strength,
			relativeStatValues.attackPower,
			relativeStatValues.hasteRating,
			relativeStatValues.hitRating,
			relativeStatValues.expertiseRating,
			relativeStatValues.agility,
			relativeStatValues.critRating);
	local maxRel = math.max(
			relativeStatValues.strength,
			relativeStatValues.attackPower,
			relativeStatValues.hasteRating,
			relativeStatValues.hitRating,
			relativeStatValues.expertiseRating,
			relativeStatValues.agility,
			relativeStatValues.critRating);

	assert(minRel);
	assert(maxRel);

	local function getStatText(hpFromStat)
		assert(hpFromStat);
		local s = Toolkit:RangeColorText(
			string.format("%.2f", hpFromStat),
			hpFromStat,
			minRel, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b, 
			maxRel, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)--]]
		--		local s = HIGHLIGHT_FONT_COLOR_CODE..string.format("%.1f", hpFromStat)..FONT_COLOR_CODE_CLOSE;
		return s;
	end;

	self:AddLine(" ");
	self:AddDoubleLine(L["Relative Stat Values"], L["MeleePoints"], 
		HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b,
		HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);

	self:AddDoubleLine("1.00 ".."Strength"..":",			getStatText(relativeStatValues.strength)); --"Strength"
	self:AddDoubleLine("1.00 ".."Attack Power"..":",		getStatText(relativeStatValues.attackPower)); --"Attack Power"
	self:AddDoubleLine("1.00 ".."Haste Rating"..":",		getStatText(relativeStatValues.hasteRating)); --"Haste Rating"
	self:AddDoubleLine("1.00 ".."Hit Rating"..":",			getStatText(relativeStatValues.hitRating)); --"Hit Rating"
	self:AddDoubleLine("1.00 ".."Expertise Rating"..":",	getStatText(relativeStatValues.expertiseRating)); --"Expertise Rating"
	self:AddDoubleLine("1.00 ".."Agility"..":",				getStatText(relativeStatValues.agility)); --"Agility"
	self:AddDoubleLine("1.00 ".."Crit Rating"..":",			getStatText(relativeStatValues.critRating)); --"Crit Rating"
	

	--[[local function GetStatText(mpFromStat)
		assert(mpFromStat);
		local s = TS.GetRangeColoredText(
				string.format("%.4f", mpFromStat),
				mpFromStat,
				minRel, 1, 1, 1, 
				maxRel, 0, 1, 0)
		return s;
	end;--]]
			
	--self:AddLine(" ");
	--self:AddLine(GREEN_FONT_COLOR_CODE..L["Hint: Click to toggle the calculator window"]..FONT_COLOR_CODE_CLOSE);
	--self:AddLine(GREEN_FONT_COLOR_CODE..L["Right-click to open the options menu"]..FONT_COLOR_CODE_CLOSE)
end;

function dataObject:GetValue(currentStats)
	assert(currentStats, "dataObject:GetValue: currentStats is nil");

	local text;
	
	local showValueNum = 1; --1=MeleePoints, 2=Combat table multiplier

	if (showValueNum == 2) then --Combat table Multiplier
		text = format("%.1f", currentStats.combatTableDamageMultiplier);
		suffix = "%";
	else --MeleePoints, and everything else
		text = format("%.1f", currentStats.meleePoints);
		suffix = " MP";
	end
     
	return text, suffix;
end

--[[
	Called by the outside world controller when the value has changed and needs to be updated
--]]
function addon:UpdateLdb()
	if (dataObject == nil) then
		return;
	end;

	local data = addon:GetCurrentMeleePoints();
	if (data == nil) then
		return
	end;

	local text, suffix = dataObject:GetValue(data);
	dataObject.value = text;
	dataObject.suffix = suffix;
end;

function addon:RegisterLDBDataObjects()
	self:UpdateLdb();
end;