local addon = MeleePoints
local profileDB --set during :SetupOptions()

function addon:SetupOptions()
	local addon = self;
	
	local AceConfig = LibStub("AceConfig-3.0");
	AceConfig:RegisterOptionsTable("MeleePoints", addon:GetOptions(), {"meleepoints", "mp"})

	-- Setup Blizzard option frames
	addon.optionsFrames = {}
	local aceConfigDialog = LibStub("AceConfigDialog-3.0");
	local rootNodeName = "MeleePoints"; --L["MeleePoints"]; --we need to use the same localized parent caption because children are placed under a parent node by caption
	addon.optionsFrames.general = aceConfigDialog:AddToBlizOptions("MeleePoints", rootNodeName,      nil,          "general") --options.args.general
	--addon.optionsFrames.mob =     aceConfigDialog:AddToBlizOptions("MeleePoints", L["Mob Stats"],    rootNodeName, "mob") --options.args.mob
	--addon.optionsFrames.player =  aceConfigDialog:AddToBlizOptions("MeleePoints", L["Player Stats"], rootNodeName, "player") --options.args.player
	
	addon.optionsFrames.general.default = function() addon:SetDefaultGeneralOptions() end;
--	addon.optionsFrames.mob.default = function() addon:SetDefaultMobStats() end;

	profileDB = addon.db.profile;
end;

local options = nil; --set on demand in GetOptions()
function addon:GetOptions()
	if (options == nil) then
		options = { 
				name = "MeleePoints",
				type = "group",
				args = {
					general = {
						type = "group",
						name = "MeleePoints Options",
						desc = "Change general MeleePoints options",
						cmdInline = true,
						args = {
							test = {
								type = "execute",
								name = "Test",
								desc = "Toggle display of debugging messages",
									func = function()
										addon:Test();
									end,
							},
							dumptable = {
								type = "execute",
								name = "Dump Table",
								desc = "Display the melee points data table",
									func = function()
										addon:DumpTable();
									end,
								
							},
							refresh = {
								type = "execute",
								name = "UpdateSourceData",
								desc = "Recalculate players melee dps value",
									func = function()
										addon:UpdateMeleePoints("Recalculate command");
										addon:Print(string.format("MeleePoints: %.1f", addon.sourceData.meleePoints));
										addon:Print(string.format("Melee DPS: %.1f", addon.sourceData.whiteDps));
									end,
							},
							dumpstatvalues = {
								type = "execute",
								name = "Dump Stat Values",
								desc = "Display the relative value of stats towards MeleePoints",
									func = function()
										addon:DumpStatValues();
									end,
								
							},
							level = {
								type = "range",
								name = "Mob Level", --L["Mob Level"],
								desc = "Sets the level difference between the mob and you", --L["Sets the level difference between the mob and you"],
								get = function(info) return profileDB.mobLevelDiff end,
								set = function(info, value)
									profileDB.mobLevelDiff = value
									addon:UpdateMeleePoints("Mob level")
									-- Update Calculator
									--if MeleePointsCalculatorFrame:IsVisible() then
									--	MPCalc:UpdateResults()
									--end
								end,
								min = 0,
								max = 3,
								step = 1,
								order = 1,
							},
							infront = {
								type = "range",
								name = "Time behind enemy", --L["Time behind enemy"],
								desc = "Sets the percentage of time you are attacking from behind.", --L["Sets teh percentage of time you are attacking from behind"],
								get = function(info) return profileDB.timeSpentBehind end,
								set = function(info, value)
									profileDB.timeSpentBehind = value
									addon:UpdateMeleePoints("Time behind enemy")
									-- Update Calculator
									--if MeleePointsCalculatorFrame:IsVisible() then
									--	MPCalc:UpdateResults()
									--end
								end,
								min = 0,
								max = 100,
								step = 1,
								order = 2,
							},
							debug = {
								name = "Debugging",
								desc = "Toggle debugging for this addon.",
								type = "toggle",
								get = function(info) return addon:IsDebugging() end,
								set = function(info, value) addon:SetDebugging(value) end,
								order = -2,
							},
							showMeleePointsDiff = {
								name = "Show Melee Points Diff",
								desc = "Toggle display of difference in tooltips.",
								type = "toggle",
								get = function(info) return addon.db.profile.showMeleePointsDiff end,
								set = function(info, value) addon.db.profile.showMeleePointsDiff = value end,
								order = 10,
							},
							showMeleePointsDiffPercent = {
								name = "Show Melee Points Diff (%)",
								desc = "Toggle display of percentage difference in tooltips.",
								type = "toggle",
								get = function(info) return addon.db.profile.showMeleePointsDiffPercent end,
								set = function(info, value) addon.db.profile.showMeleePointsDiffPercent = value end,
								order = 11,
							}

						},
					},
				},
			}
	end;
	
	return options;

end

function addon:SetDefaultGeneralOptions()
	profileDB.mobLevelDiff = 0;
	profileDB.timeSpentBehind = 90;
	profileDB.showMeleePointsDiff = true;
	profileDB.showMeleePointsDiffPercent = true;
	addon:UpdateMeleePoints("SetDefaultGeneralOptions");

	LibStub("AceConfigRegistry-3.0"):NotifyChange("MeleePoints");
	
	self:Print("General options reset to default");
end;

function addon:ShowConfig()
	--called by /tp config console option
	
	-- Open a child item first, so the menu expands
	--InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.mob) --a child item
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.general) --now show the real thing we want to show
end