﻿---------------
-- Libraries --
---------------
local TipHooker = LibStub:GetLibrary("LibTipHooker-1.1")
local StatLogic = LibStub:GetLibrary("LibStatLogic-1.2")
--local L = LibStub("AceLocale-3.0"):GetLocale("MeleePoints") --Get the localization for our addon
local ReforgingInfo = LibStub:GetLibrary("LibReforgingInfo-1.0");


--------------------
-- Initialization --
--------------------
MeleePointsTooltips = {}
local tips = MeleePointsTooltips
local MP = MeleePoints

-- Color Numbers
local GREEN_FONT_COLOR_CODE = GREEN_FONT_COLOR_CODE -- "|cff20ff20" Green
local HIGHLIGHT_FONT_COLOR_CODE = HIGHLIGHT_FONT_COLOR_CODE -- "|cffffffff" White
local RED_FONT_COLOR_CODE = RED_FONT_COLOR_CODE -- "|cffffffff" White
local FONT_COLOR_CODE_CLOSE = FONT_COLOR_CODE_CLOSE -- "|r"
local function colorNum(text, num)
	if num > 0 then
		return GREEN_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE
	elseif num < 0 then
		return RED_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE
	else
		return HIGHLIGHT_FONT_COLOR_CODE..text..FONT_COLOR_CODE_CLOSE
	end
end

function MeleePointsTooltips:OnEnable()
	--Called by MeleePoints when MeleePoints addon is enabled (OnEnable)
	
	-- Hook item tooltips
	--MP:Debug("Hooking MeleePoints tooltips");
	TipHooker:Hook(self.ProcessTooltip, "item")
end;

function MeleePointsTooltips:OnDisable()
	--Called by MeleePoints when MeleePoints addon is disabled (OnDisable)
	-- Unhook item tooltips
	--MP:Debug("Unhooking MeleePoints tooltips");
	TipHooker:Unhook(self.ProcessTooltip, "item")
end;

--[[
	Callback function given to TipHooker
	@param tooltip The blizzard tooltip frame, supporting all the methods it supports, e.g.
			AddDoubleLine("textLeft", "textRight")
			AddLine("text")
			Show()
		See http://wowprogramming.com/docs/widgets/GameTooltip
	@param name (string) The name of the item we're providing the tooltip for (e.g. "Lavascale Fillet")
	@param link (string,link) The string in-game link of the item we're getting the tooltip for (e.g. "[Lavascale Fillet]")
--]]
function MeleePointsTooltips.ProcessTooltip(tooltip, name, link, ...)
	-- Check if we're in standby mode
	if not MeleePoints:IsEnabled() then return end
	
	assert(tooltip);
	
	-- Get diff tables
	local diffTable1, diffTable2
	--if MeleePointsTooltips:GetItemSubType(link) == L["Gems"] then
--		diffTable1 = StatLogic:GetSum(link)
--	else
		diffTable1, diffTable2 = StatLogic:GetDiff(link)
	--end
	
	-- Check for bags
	if not diffTable1 then
		return
	end
	
	
	
	if IsShiftKeyDown() then
		--Add the itemLink if they're pressing Shift
		--tooltip:AddLine(string.gsub(link, "|", "||")); --escape | to ||
		MP:Debug("DiffTable:");
		MP:Debug(MP:VarAsString(diffTable1));
	end;

	local changes = MeleePointsTooltips:BuildChanges(nil, diffTable1);

	if IsShiftKeyDown() and IsControlKeyDown() then
		MP:Print("Changes:");
		MP:Print(MP:VarAsString(changes));
	end;

	--MeleePoints:Debug("ProcessTooltip(tooltip, name, link, ...)");
	if (MP.db.profile.showMeleePointsDiff or MP.db.profile.showMeleePointsDiffPercent) then
		--Get the apply changes and get new MeleePoints
		local target = {};
		copyTable(target, MP.sourceData);
		MP:ApplyChanges(target, changes);

		local diff = target.meleePoints-MP.sourceData.meleePoints;
	
		local szPoints = "";
		if (MP.db.profile.showMeleePointsDiff) then
			szPoints = szPoints..colorNum(string.format("%+.1f", diff), diff);
		end;
		
		if (MP.db.profile.showMeleePointsDiffPercent) then
			diff = diff / MP.sourceData.meleePoints * 100;
			local diffString = colorNum(string.format("%+.1f%%", diff), diff);
			if (szPoints ~= "") then
				szPoints = szPoints.." ("..diffString..")";
			else
				szPoints = szPoints..diffString;
			end;
		end;
		tooltip:AddDoubleLine("MeleePoints:", szPoints);
	end;

	if (ReforgingInfo) then
		if ReforgingInfo:IsItemReforged(link) then
			local minus, plus = ReforgingInfo:GetReforgedStatNames(ReforgingInfo:GetReforgeID(link))
			tooltip:AddDoubleLine("Reforged:", string.format("%s%s%s -> %s%s%s", 
					RED_FONT_COLOR_CODE, minus, FONT_COLOR_CODE_CLOSE, 
					GREEN_FONT_COLOR_CODE, plus, FONT_COLOR_CODE_CLOSE))
		end
	end
	
	--The tooltip is not automatically resized to fit the added line; to do so, call the tooltip's :Show() method after adding lines.	
	tooltip:Show(); 
end;

--[[
	MeleePointsTooltips:GetItemSubType(itemID|"name"|"itemlink")
	Returns a localized subclass name of an item (e.g. "One-Handed Swords", "Daggers", "Wands") 
	(see GetAuctionItemSubClasses for complete list)
--]]
function MeleePointsTooltips:GetItemSubType(item)
	local _, _, _, _, _, _, itemSubType = GetItemInfo(item)
	return itemSubType
end

-- Build changes table for MeleePoints:ApplyChanges from StatLogic:GetDiff table
function MeleePointsTooltips:BuildChanges(reserved, diffTable)
	--MP:Debug("TPTips:BuildChanges()  StatLogic changes diffTable="..MP:VarAsString(diffTable))

	assert(diffTable);
	
	--MP:Print(MP:VarAsString(diffTable));
	--MP:Print(MP:VarAsString(changes));

	
	local changes = {}
	
	changes.hasteRating = diffTable.MELEE_HASTE_RATING;
	changes.haste = diffTable.MELEE_HASTE;
	changes.strength = diffTable.STR;
	--changes.attackPower = 0;
	
	changes.agility = diffTable.AGI;
	changes.stamina = diffTable.STA;
	
	changes.hitRating = diffTable.MELEE_HIT_RATING;
	
	changes.expertiseRating = diffTable.EXPERTISE_RATING;
	changes.masteryRating = diffTable.MASTERY_RATING;
	
	--weapon
	if (diffTable.DPS or diffTable.MAX_DAMAGE or diffTable.MIN_DAMAGE) then
		if (diffTable.MIN_DAMAGE) then
			changes.weaponDamageMinBase = MIN_DAMAGE;
		end;
		if (diffTable.MAX_DAMAGE) then
			changes.weaponDamageMaxBase = MAX_DAMAGE;
		end;
		if (diffTable.DPS) then
			changes.weaponSpeedBase = diffTable.ATTACK_SPEED;
		end;
	end;
	--[[
		Future potentially useful diff stats:
		
		IGNORE_ARMOR
		AP
		RANGED_AP
		FERAL_AP
		
		RANGED_HIT_RATING
		
		RANGED_HASTE_RATING
		MASTERY_RATING
		ARMOR_PENETRATION_RATING
		MELEE_HIT
		RANGED_HIT
		MELEE_CRIT
		RANGED_CRIT
		
		RANGED_HASTE
		EXPERTISE
		ARMOR_PENETRATION
		MASTERY
		MELEE_CRIT_DMG
		RANGED_CRIT_DMG
		CAST_TIME
			--]]
	--[[
	changes.playerHealth = diffTable.HEALTH;
	changes.armorFromItems = diffTable.ARMOR;
	changes.armor = diffTable.ARMOR_BONUS;
	changes.defenseRating = diffTable.DEFENSE_RATING;
	if diffTable.DODGE_RATING then
		changes.dodgeChance = StatLogic:GetAvoidanceGainAfterDR("DODGE", StatLogic:GetEffectFromRating(diffTable.DODGE_RATING, CR_DODGE, MP.playerLevel)) * 0.01
	end
	if diffTable.PARRY_RATING then
		changes.parryChance = StatLogic:GetAvoidanceGainAfterDR("PARRY", StatLogic:GetEffectFromRating(diffTable.PARRY_RATING, CR_PARRY, MP.playerLevel)) * 0.01
	end
	if diffTable.BLOCK_RATING then
		changes.blockChance = StatLogic:GetEffectFromRating(diffTable.BLOCK_RATING, CR_BLOCK, MP.playerLevel) * 0.01
	end
	--]]
	if diffTable.MASTERY_RATING then
		changes.masteryRating = diffTable.MASTERY_RATING;
	end

	--changes.blockValue = diffTable.BLOCK_VALUE
	--changes.resilience = diffTable.RESILIENCE_RATING
	return changes
end
